; ---purpose---
; Evaluate arithmetic expressions.
; Evaluate all operations with numeric arguments
; (or subexpressions evaluating to numeric arguments)
; to their results. Reduce expressions with one variable
; argument using rules such as X+0 -> X and X^0 -> 1.
; Return the evaluated expression.
; --
; The following functions are recognized:
; IPLUS, IDIFFERENCE, ITIMES, IQUOTIENT, IEXPT, MINUS.
; Variables are represented by single letters. Numbers
; are represented by quoted lists of digits.

; ---keywords---
; EVALUATE function, evaluation, partial, math expressions
; arithmetic expressions, arithmetics

; ---args---
; X - expression

; ---example---
; (EVALUATE '(ITIMES X (IEXPT (QUOTE (+ 2))
;       (IPLUS (QUOTE (+ 3)) (QUOTE (+ 4))))))

:L LIB/MEMBER.L
:L LIB/CAAR.L
;L LIB/CADR.L
:L LIB/CAAAR.L
;L LIB/CADDR.L
:L LIB/LOGOP.L
;L LIB/LOGNOT.L
;L LIB/LOGAND.L
:L LIB/ZEROP.L
:L LIB/IPLUS.L
:L LIB/IDIFFERENCE.L
:L LIB/ITIMES.L
:L LIB/IQUOTIENT.L
:L LIB/IEXPT.L
:L LIB/MINUS.L

; ---code---
(DEFUN EVALUATE (X)
	(LABEL

    ((SYMBOLP (LAMBDA (X)
		(MEMBER X '#ABCDEFGHIJKLMNOPQRSTUVWXYZ)))

	(NUMERICP (LAMBDA (X)
		(COND ((ATOM X) *F*) (T (EQ (CAR X) 'QUOTE)))))

	; Evaluate (fn left-num right-num)
	(COMPUTE (LAMBDA (OP LEFT RIGHT)
		(COND ((EQ (CAR X) 'IPLUS) (IPLUS LEFT RIGHT))
			((EQ (CAR X) 'IDIFFERENCE) (IDIFFERENCE LEFT RIGHT))
			((EQ (CAR X) 'ITIMES) (ITIMES LEFT RIGHT))
			((EQ (CAR X) 'IQUOTIENT) (IQUOTIENT LEFT RIGHT))
			((EQ (CAR X) 'IEXPT) (IEXPT LEFT RIGHT))
			(T (BOTTOM 'COMPUTE)))))

	(ZERO-P (LAMBDA (X)
		(COND ((LOGNOT (NUMERICP X)) *F*)
			(T (ZEROP (CADR X))))))

	(ONE-P (LAMBDA (X)
		(COND ((LOGNOT (NUMERICP X)) *F*)
			(T (EQUAL X ''#+1)))))

	(EVL (LAMBDA (X)
		(COND ((SYMBOLP X) X)	; Pass thru symbols and numbers
			((NUMERICP X) X)
			; Evaluate MINUS with constant argument
			((EQ (CAR X) 'MINUS)
				(LABEL ((FAC (EVL (CADR X))))
					(COND ((NUMERICP FAC)
							(LIST 'QUOTE (MINUS (CADR FAC))))
						(T (LIST 'MINUS FAC)))))
			; Evaluate or reduce binary functions
			((MEMBER (CAR X) '(IPLUS IDIFFERENCE ITIMES IQUOTIENT IEXPT))
				; First evaluate the arguments
				(LABEL ((LEFT (EVL (CADR X)))
						(RIGHT (EVL (CADDR X))))
					(COND
						; Both arguments constant, compute the result
						((LOGAND (NUMERICP LEFT) (NUMERICP RIGHT))
							(LIST QUOTE (COMPUTE (CAR X) (CADR LEFT)
								(CADR RIGHT))))
						((EQ (CAR X) 'IPLUS)
							(COND ((ZERO-P RIGHT) LEFT)		; X+0 => X
								((ZERO-P LEFT) RIGHT)		; 0+X => X
								(T (LIST (CAR X) LEFT RIGHT))))
						((EQ (CAR X) 'IDIFFERENCE)
							(COND ((ZERO-P RIGHT) LEFT)		; X-0 => X
								((ZERO-P LEFT)				; 0-X => -X
									(LIST 'MINUS RIGHT))
								(T (LIST (CAR X) LEFT RIGHT))))
						((EQ (CAR X) 'ITIMES)
							(COND ((ZERO-P RIGHT) ''#0)		; X*0 => 0
								((ZERO-P LEFT) ''#0)		; 0*X => 0
								((ONE-P LEFT) RIGHT)		; X*1 => X
								((ONE-P RIGHT) LEFT)		; 1*X => X
								(T (LIST (CAR X) LEFT RIGHT))))
						((EQ (CAR X) 'IQUOTIENT)
							(COND ((ZERO-P RIGHT)			; X/0 => _|_
									(BOTTOM 'DIVIDE-BY-ZERO))
								((ZERO-P LEFT) ''#0)		; 0/X => 0
								((ONE-P RIGHT) LEFT)		; X/1 => X
								(T (LIST (CAR X) LEFT RIGHT))))
						((EQ (CAR X) 'IEXPT)
							(COND ((ZERO-P RIGHT) ''#1)		; X^0 => 1
								((ZERO-P LEFT) ''#0)		; 0^X => 0
								((ONE-P LEFT) ''#1)			; 1^X => 1
								((ONE-P RIGHT) LEFT)		; X^1 => X
								(T (LIST (CAR X) LEFT RIGHT))))
						; Unknown operator, pass thru
						(T (LIST (CAR X) LEFT RIGHT)))))
			(T (BOTTOM (LIST EVL: X)))))))

	(EVL X)))

