; ---purpose---
; Convert arithmetic expressions from prefix to
; infix notation. Prefix expressions may contain
; variables [symbols like X], numbers [quoted
; integer numbers like '#+57], and these functions:
; IPLUS, IDIFFERENCE, ITIMES, IQUOTIENT, IEXPT,
; MINUS. Functions are translated to these
; operators: + - * / ^.
; --
; For example,
; -
; (INFIX '(IPLUS '#+15 '#+17))
; -
; gives
; -
; '(#15 + #17).
; --
; INFIX will insert parentheses ([ and ]) where
; required.

; ---keywords---
; INFIX function, prefix to infix, RPN, conversion

; ---args---
; X - arithmetic expression

; ---example---
; (INFIX '(IPLUS '#+2 (ITIMES '#+3 (IEXPT '#+4 (IPLUS '#+5 '#+6)))))

:L LIB/DIGITS.L
:L LIB/CAAR.L
;L LIB/CADR.L
:L LIB/CAAAR.L
;L LIB/CAADR.L
;L LIB/CADDR.L
;L LIB/CDADR.L
:L LIB/CAAAAR.L
;L LIB/CAADDR.L
:L LIB/APPEND.L
:L LIB/MEMBER.L
:L LIB/ASSOC.L
:L LIB/IPLUS.L
:L LIB/IDIFFERENCE.L
:L LIB/ITIMES.L
:L LIB/IQUOTIENT.L
:L LIB/IEXPT.L

; ---code---
(DEFUN INFIX (X)
	(LABEL

	; Operator symbols
	((+ '+) (- '-) (* '*) (/ '/) (^ '^) ([ '[) (] '])

	; Function->operator translation
	(OPS '((IPLUS.+) (IDIFFERENCE.-) (ITIMES.*) (IQUOTIENT./) (IEXPT.^)))

	; Precedence list (descending order)
	(PRECEDENCE '((HIGH) (IEXPT) (ITIMES IQUOTIENT) (IPLUS IDIFFERENCE) (LOW)))

	(FUNCTIONP (LAMBDA (X)
		(MEMBER X '(IPLUS IDIFFERENCE ITIMES IQUOTIENT IEXPT))))

	(SYMBOLP (LAMBDA (X)
		(MEMBER X '#ABCDEFGHIJKLMNOPQRSTUVWXYZ)))

	; Numbers are always quoted, like (QUOTE #+5)
	(NUMERICP (LAMBDA (X)
		(COND ((ATOM X) *F*) (T (EQ (CAR X) 'QUOTE)))))

	; Has the function associated with the operator X
	; a higher precedence than Y?
	(HIGHERPRECP (LAMBDA (X Y)
		(LABEL ((HP (LAMBDA (X Y L)
			(COND ((NULL L) *F*)
				((MEMBER X (CAR L))
					(COND ((MEMBER Y (CAR L)) *F*) (T *T*)))
				((MEMBER Y (CAR L)) *F*)
				(T (HP X Y (CDR L)))))))
		(HP X Y PRECEDENCE))))

	; Is X the application of a function?
	(OPERATIONP (LAMBDA (X)
		(COND ((ATOM X) *F*)
			((FUNCTIONP (CAR X)) *T*)
			(T *F*))))

	(APPEND3 (LAMBDA (A B C) (APPEND A (APPEND B C))))

	(PARENTHESIZE (LAMBDA (X) (APPEND3 '#[ X '#])))

	(INF (LAMBDA (X OP)
		(COND ((NUMERICP X)
				; Number: remove positive sign
				(COND ((EQ (CAADR X) -) (CADR X))
					(T (CDADR X))))
			((SYMBOLP X) (LIST X))
			((EQ (CAR X) 'MINUS)
				(APPEND (LIST -) (INF (CADR X) 'HIGH)))
			; Binary function?
			((FUNCTIONP (CAR X))
				; Convert prefix to infix
				(LABEL ((SUB (APPEND3 (INF (CADR X) (CAR X))
						(LIST (CDR (ASSOC (CAR X) OPS)))
						(INF (CADDR X) (CAR X)))))
					; If the surrounding operation has a higher
					; precedence, parenthesize this expression
					(COND ((HIGHERPRECP OP (CAR X))
							(PARENTHESIZE SUB))
						(T SUB))))
			(T (BOTTOM (LIST 'SYNTAX: X)))))))

	(INF X 'LOW)))

